from enum import Enum
from typing import Optional, Dict

from pydantic import BaseModel


class SinkOperationType(Enum):
    UPSERT = 1
    DELETE = 2
    INSERT = 4


class SinkOperation(BaseModel):
    sink_operation_type: SinkOperationType

    # update query for update by query
    update_query: Dict = {}

    # source field name for script-based update
    source_field_name: Optional[str]

    # this can be a json or any primitive value
    new_val: object = None

    class Config:
        arbitrary_types_allowed = True


class SinkRecordDTO(BaseModel):
    sink_operation: SinkOperation
    message: Dict
    key: Optional[str]
    offset: Optional[int]
    topic: Optional[str]
    partition: Optional[int]


class DeadLetterDTO(BaseModel):
    key: str
    message: str
    topic: str
    partition: int
    failed_at: str
    error: str
    offset: int



import json
from typing import TypeVar, Generic, List

from kafka import KafkaProducer

from src.exceptions.usi_exceptions import BadConsumerConfigException
from src.kafka_core.ser_des_util import get_ser_des
from src.stream_writers.stream_writer import StreamWriter
from src.utility import logging_util
from src.utility.common_util import CLIENT_ID

logger = logging_util.get_logger(__name__)

T = TypeVar('T')


class KafkaStreamWriter(Generic[T], StreamWriter):

    def __init__(self, config: dict):
        self.config = config
        self.kafka_producer = self.__create_kafka_producer()
        if not config.get('topic_name'):
            raise BadConsumerConfigException('missing producer topic name.')
        self.topic = config.get('topic_name')

    def write(self, streams: List[T]) -> None:
        """
        writes message of type T to a kafka topic
        :param streams: list of messages
        :return: None
        """
        for event in streams:
            key = None
            if hasattr(event, 'key'):
                key = event.key

            self.kafka_producer.send(topic=self.topic, key=key, value=json.dumps(event.__dict__))

    def close(self) -> None:
        """
        closes the writer kafka producer object
        :return: None
        """
        if self.kafka_producer:
            self.kafka_producer.close()

    def __create_kafka_producer(self) -> KafkaProducer:
        bootstrap_servers = self.config.get('bootstrap_servers')

        return KafkaProducer(bootstrap_servers=bootstrap_servers,
                             key_serializer=get_ser_des(self.config.get(
                                 'key_serializer', 'STRING_SER')),
                             value_serializer=get_ser_des(self.config.get(
                                 'value_serializer', 'STRING_SER')),
                             acks=self.config.get('acks', 'all'),
                             compression_type=self.config.get('compression_type',
                                                              'gzip'),
                             retries=self.config.get('retries', 1),
                             linger_ms=self.config.get('linger_ms', 10),
                             client_id=CLIENT_ID)



import logging
import os

LOG_FILE_PATH = '/var/www-api/.log/consumer_app.log'
LOG_DIR_PATH = '/var/www-api/.log'
LOG_FORMATTER = '%(levelname)s - %(asctime)s - %(name)s - %(message)s'


def is_valid_path(path: str):
    if os.path.exists(path) or os.access(os.path.dirname(path), os.W_OK):
        return True
    return False


def get_logger(name=None):
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    std_out_handler = StdOutHandler()
    std_out_handler.setLevel(logging.INFO)
    logger.addHandler(std_out_handler)

    if is_valid_path(LOG_DIR_PATH):
        file_out_handler = FileOutHandler()
        file_out_handler.setLevel(logging.INFO)
        logger.addHandler(file_out_handler)

    return logger


class StdOutHandler(logging.StreamHandler):
    def __init__(self, stream=None):
        super(StdOutHandler, self).__init__()

    def format(self, record):
        self.formatter = logging.Formatter(LOG_FORMATTER)
        return super(StdOutHandler, self).format(record)


class FileOutHandler(logging.FileHandler):
    def __init__(self):
        super(FileOutHandler, self).__init__(filename=LOG_FILE_PATH)

    def format(self, record):
        self.formatter = logging.Formatter(LOG_FORMATTER)
        return super(FileOutHandler, self).format(record)



import logging
import time
from abc import ABC
from typing import List

from kafka.consumer.fetcher import ConsumerRecord
from ratelimit import limits, sleep_and_retry

from src.exceptions.usi_exceptions import BadConsumerConfigException
from src.kafka_core.kafka_stream_writer import KafkaStreamWriter
from src.model.worker_dto import DeadLetterDTO, SinkRecordDTO
from src.stream_writers.stream_writer import StreamWriter, get_stream_writers
from src.transformers.transformer import get_transformer

ONE_SECOND = 1
CALLS = 20


class SinkTask(ABC):

    def __init__(self, config: dict):
        self.sink_configs = config.get('sink_configs')
        if self.sink_configs is None:
            raise BadConsumerConfigException('Missing Sink Config.')
        self.config = config
        processor_cls_path = self.sink_configs.get('transformer_cls')
        if not processor_cls_path:
            raise BadConsumerConfigException('sink_configs.transformer_cls is a mandatory config')
        self.stream_transformer = get_transformer(processor_cls_path, self.sink_configs)
        self.operation_extractor = None
        stream_writer_cls_paths: List[str] = self.sink_configs.get('stream_writers')
        if not stream_writer_cls_paths or len(stream_writer_cls_paths) == 0:
            raise BadConsumerConfigException('sink_configs.stream_writers is a mandatory config')
        self.sink_stream_writers: List[StreamWriter] = get_stream_writers(
            stream_writer_cls_paths, self.sink_configs)
        if config.get('dlq_config') is not None:
            self.dlq_stream_writer: KafkaStreamWriter[DeadLetterDTO] = KafkaStreamWriter(
                config.get('dlq_config'))
        self.retries = self.sink_configs.get('num_retries', 3)
        self.retry_delay_seconds = self.sink_configs.get('retry_delay_seconds', 1)

    def write_to_sink(self, sink_record_dto_list: List[SinkRecordDTO]):
        for stream_writer in self.sink_stream_writers:
            retries = 0
            while retries <= self.retries:
                try:
                    stream_writer.write(sink_record_dto_list)
                    break
                except Exception as e:
                    if retries == self.retries:
                        raise e
                    retries = retries + 1
                    logging.error(f'{type(stream_writer)} - Failed with exception: {e}, retrying '
                                  f'attempt'
                                  f' {retries}')
                    time.sleep(self.retry_delay_seconds)

    @sleep_and_retry
    @limits(calls=CALLS, period=1)
    def process(self, consumer_records: List[ConsumerRecord]):

        for consumer_record in consumer_records:
            try:
                sink_record_dto: SinkRecordDTO = self.stream_transformer.transform(consumer_record)
                sink_record_dto_list: List[SinkRecordDTO] = [sink_record_dto]
            except Exception as e:
                self.handle_dlq_push(consumer_record.key, consumer_record.value,
                                     consumer_record.topic, consumer_record.partition,
                                     'TRANSFORM', e, consumer_record.offset)
                continue

            try:
                self.write_to_sink(sink_record_dto_list)
            except Exception as e:
                self.handle_dlq_push(consumer_record.key, consumer_record.value,
                                     consumer_record.topic, consumer_record.partition,
                                     'SINK_UPDATE', e, consumer_record.offset)

    def handle_dlq_push(self, key: str, message: str, topic: str, partition: int,
                        failed_at: str, error: Exception, offset: int):
        logging.warning(
            f'failed to {failed_at} key: {key} and message: {message}, in topic {topic} '
            f'having offset {offset}, with error: {error}')
        try:
            if self.dlq_stream_writer is not None:
                dead_letter = DeadLetterDTO(key=key, message=message, topic=topic,
                                            partition=partition, failed_at=failed_at,
                                            error=str(error) if error is not None else "",
                                            offset=offset)
                self.dlq_stream_writer.write([dead_letter])
        except Exception as e:
            logging.error(f'Failed to write to DLQ: {e}')



import uuid

CLIENT_ID = str(uuid.uuid4())


def singleton(cls):
    instances = {}

    def get_instance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]

    return get_instance



import logging
from typing import List

from src.model.worker_dto import SinkRecordDTO
from src.stream_writers.stream_writer import StreamWriter


class ConsoleStreamWriter(StreamWriter):

    def write(self, streams: List[SinkRecordDTO]) -> None:
        """
        Writes processed records read from kafka to Elastic search
        :param streams: List of SinkRecordDTO - transformed data to be written to ES
        :return: None
        """
        for sink_record_dto in streams:
            logging.info(f' Key: {sink_record_dto.key} - value: {sink_record_dto.message}')

    def close(self) -> None:
        pass



import importlib
from abc import ABC, abstractmethod

from kafka.consumer.fetcher import ConsumerRecord

from src.exceptions.usi_exceptions import BadInput
from src.model.worker_dto import SinkRecordDTO


class StreamTransformer(ABC):

    def __init__(self, config: dict):
        self.config = config

    @abstractmethod
    def transform(self, consumer_record: ConsumerRecord) -> SinkRecordDTO:
        """
        Transforms the JSON for sink updated and extracts the operation associated with the event
        :param consumer_record: kafka consumer record
        :return: return the Sink record that will be put into sink Datastore
        """


def get_transformer(cls_path: str, config: dict) -> StreamTransformer:
    module_name, class_name = cls_path.rsplit(".", 1)
    stream_transformer = getattr(importlib.import_module(module_name), class_name)

    if not issubclass(stream_transformer, StreamTransformer):
        raise BadInput(f'{cls_path} is not a subclass of StreamTransformer')

    return stream_transformer(config)



import secrets
import time

from fastapi import FastAPI, Depends, HTTPException, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from starlette.requests import Request
from starlette.responses import JSONResponse

from src import USERNAME, PASSWORD
from src.exceptions.usi_exceptions import BadInput, GenericException
from src.kafka_core.consumer_manager import ConsumerWorkerManager
from src.utility import logging_util

logger = logging_util.get_logger(__name__)

app = FastAPI(title="Distributed Kafka Consumer Using Ray - Manager")
security = HTTPBasic()
cwm = ConsumerWorkerManager()


def authorize(credentials: HTTPBasicCredentials = Depends(security)):
    correct_username = secrets.compare_digest(credentials.username, USERNAME)
    correct_password = secrets.compare_digest(credentials.password, PASSWORD)
    if not (correct_username and correct_password):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Basic"},
        )
    return credentials.username


@app.on_event("startup")
def on_startup():
    cwm.start_all_workers()


@app.on_event("shutdown")
def on_shutdown():
    cwm.start_all_workers()


@app.post('/manager/health', include_in_schema=False)
def health():
    return {'message': 'App is up!'}


@app.post('/manager/start-consumers', dependencies=[Depends(authorize)])
def start_consumers():
    cwm.start_all_workers()
    return "Successfully started all workers!"


@app.get('/manager/fetch-consumers', dependencies=[Depends(authorize)])
def get_consumers():
    return cwm.get_all_running_consumer()


@app.post('/manager/read-from-timestamp', dependencies=[Depends(authorize)])
def read_from_timestamp(consumer_name: str, start_timestamp: int,
                        end_timestamp: int = int(time.time() * 1000),
                        stop_running_consumer: bool = True):
    cwm.start_worker_with_timestamp(start_timestamp=start_timestamp, end_timestamp=end_timestamp,
                                    stop_regular=stop_running_consumer, name=consumer_name)
    return "Successfully started!"


@app.post('/manager/start-consumer/{consumer_name}', dependencies=[Depends(authorize)])
def start_consumer(consumer_name):
    cwm.start_worker(consumer_name)
    return "Successfully started worker!"


@app.post('/manager/stop-consumers', dependencies=[Depends(authorize)])
def stop_consumers():
    cwm.stop_all_workers()
    return "Successfully Stopped all workers!"


@app.post('/manager/stop-consumer/{consumer_name}', dependencies=[Depends(authorize)])
def stop_consumer(consumer_name):
    cwm.stop_worker(consumer_name)
    return "Successfully Stopped!"


@app.exception_handler(Exception)
def generic_exception_handler(request: Request, exc: Exception):
    logger.error(exc)
    return JSONResponse(
        status_code=500,
        content={"message": "Oops! I messed up!"},
    )


@app.exception_handler(GenericException)
def request_validation_exception_handler(request: Request, exc: GenericException):
    logger.error(exc)
    return JSONResponse(
        status_code=500,
        content={"message": exc.message},
    )


@app.exception_handler(BadInput)
def request_validation_exception_handler(request: Request, exc: BadInput):
    logger.error(exc)
    return JSONResponse(
        status_code=422,
        content={"message": exc.message},
    )



import copy
import json
import os

from src import WORKER_CONFIG_PATH, ROOT_SRC_DIR
from src.exceptions.usi_exceptions import BadInput
from src.utility.common_util import singleton

ENV = os.environ.get('env', 'dev')
AWS_REGION = os.environ.get('aws_region', '')


@singleton
class ConfigManager:

    def __init__(self):
        self._worker_config = None
        self._load_consumer_config()

    def _load_consumer_config(self) -> None:
        with open(ROOT_SRC_DIR + WORKER_CONFIG_PATH, "r") as f:
            self._worker_config = json.load(f)

    def get_worker_config(self) -> list:
        return copy.deepcopy(self._worker_config)

    def get_worker_config_by_name(self, name: str) -> dict:

        for config in self._worker_config:
            if config['consumer_name'] == name:
                return config

        raise BadInput(f'Consumer name: {name}, is not configured.')



from kafka import KafkaConsumer

TWO_MINUTES = 2


def is_end_offset_none(end_offsets: dict, start_offsets: dict) -> bool:
    """
    Utility function to check if the partition that has start offset has end offset too.
    :param end_offsets: topic partition and end offsets
    :param start_offsets:topic partition and start offsets
    :return: True/False
    """
    if len(end_offsets) == 0:
        return True

    for tp, offset in end_offsets.items():
        if offset is None and start_offsets[tp] is not None:
            return True

    return False


def is_all_end_offset_found(end_offsets: dict, start_offsets: dict) -> bool:
    """
    Utility function to check if the partition that has start offset has end offset too.
    :param end_offsets: topic partition and end offsets
    :param start_offsets:topic partition and start offsets
    :return: True/False
    """
    if len(end_offsets) == 0:
        return False

    for tp, offset in end_offsets.items():
        if offset is None and start_offsets[tp] is not None:
            return False

    return True


def get_start_end_offsets(start_timestamp: int, end_timestamp: int,
                          topic_partitions: set, consumer: KafkaConsumer):
    """
    Get start and end offset for all the partitions based on the given start and end timestamp
    :param start_timestamp: start timestamp in epoch time millis
    :param end_timestamp: end timestamp in epoch time millis
    :param topic_partitions: topic partition set
    :param consumer: kafka consumer
    :return: tuple of start offsets and end offsets for each partition
    """
    tp_start_timestamps: dict = {}
    for tp in topic_partitions:
        tp_start_timestamps[tp] = start_timestamp

    start_offsets = consumer.offsets_for_times(tp_start_timestamps)
    end_offsets = {}
    # go back 2 minute and keep checking if there are end offsets in partition
    tp_end_timestamps: dict = {}
    while not is_all_end_offset_found(start_offsets=start_offsets, end_offsets=end_offsets):
        for tp in topic_partitions:
            # seek previous offset from a partition only if the offset is not found
            if len(end_offsets) == 0 or (end_offsets[tp] is None and start_offsets[tp] is not
                                         None):
                tp_end_timestamps[tp] = end_timestamp

        end_offsets = consumer.offsets_for_times(tp_end_timestamps)
        end_timestamp = end_timestamp - (TWO_MINUTES * 60 * 1000)

    return start_offsets, end_offsets



import json

from kafka.consumer.fetcher import ConsumerRecord

from src.model.worker_dto import SinkRecordDTO, SinkOperation, SinkOperationType
from src.transformers.transformer import StreamTransformer


class SampleTransformer(StreamTransformer):
    def transform(self, consumer_record: ConsumerRecord) -> SinkRecordDTO:
        """
        converts message to message dict
        :param consumer_record: kafka consumer record
        :return: SinkRecordDTO
        """
        # do something here
        message_dict: dict = json.loads(consumer_record.value)
        sink_operation = SinkOperation(
            sink_operation_type=SinkOperationType.UPSERT
        )

        return SinkRecordDTO(key=consumer_record.key,
                             message=message_dict,
                             topic=consumer_record.topic,
                             offset=consumer_record.offset,
                             sink_operation=sink_operation,
                             partition=consumer_record.partition)



import importlib
from abc import abstractmethod, ABC
from typing import List

from src.exceptions.usi_exceptions import BadInput


class StreamWriter(ABC):

    def __init__(self, config: dict):
        self.config = config

    @abstractmethod
    def write(self, streams: list) -> None:
        """
        Implement this interface to create an instance of stream writer
        :param streams: key and message dictionary
        :return: None
        """

    @abstractmethod
    def close(self) -> None:
        """
        tear down
        :return:
        """


def get_stream_writers(cls_paths: List[str], config: dict) -> List[StreamWriter]:
    stream_writers: List[StreamWriter] = []
    for cls_path in cls_paths:
        module_name, class_name = cls_path.rsplit(".", 1)
        stream_writer_cls = getattr(importlib.import_module(module_name), class_name)

        if not issubclass(stream_writer_cls, StreamWriter):
            raise BadInput(f'{cls_path} is not a subclass of StreamTransformer')

        stream_writers.append(stream_writer_cls(config))
    return stream_writers
